# sage_setup: distribution = sagemath-groups
# sage.doctest: needs sage.geometry.polyhedron sage.groups
r"""
PALP normal form of vertices of a lattice polytope
"""
# ****************************************************************************
#       Copyright (C) 2013      Jan Keitel
#                     2014      Volker Braun
#                     2018      Christian Stump
#                     2019      Vincent Delecroix
#                     2019      Jonathan Kliem
#                     2021      Michael Orlitzky
#                     2018-2022 Frédéric Chapoton
#                     2023      Luze Xu
#                     2023      Matthias Koeppe
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.groups.perm_gps.permgroup_element cimport PermutationGroupElement
from sage.groups.perm_gps.permgroup_named import SymmetricGroup
from sage.matrix.matrix_integer_dense cimport Matrix_integer_dense
from sage.matrix.special import column_matrix
from sage.structure.element import Matrix


def _palp_PM_max(Matrix_integer_dense PM, check=False):
    r"""
    Compute the permutation normal form of the vertex facet pairing matrix.

    The permutation normal form of a matrix is defined as the lexicographic
    maximum under all permutations of its rows and columns. For more
    more detail, see also
    :meth:`~sage.matrix.matrix2.Matrix.permutation_normal_form`.

    Instead of using the generic method for computing the permutation
    normal form, this method uses the PALP algorithm to compute
    the permutation normal form and its automorphisms concurrently.

    INPUT:

    - ``check`` -- boolean (default: ``False``); whether to return
      the permutations leaving the maximal vertex-facet pairing
      matrix invariant

    OUTPUT:

    A matrix or a tuple of a matrix and a dict whose values are the
    permutation group elements corresponding to the permutations
    that permute :meth:`vertices` such that the vertex-facet pairing
    matrix is maximal.

    EXAMPLES::

        sage: o = lattice_polytope.cross_polytope(2)
        sage: PM = o.vertex_facet_pairing_matrix()
        sage: PM_max = PM.permutation_normal_form()
        sage: PM_max == o._palp_PM_max()
        True

        sage: # needs polytopes_db
        sage: P2 = ReflexivePolytope(2, 0)
        sage: PM_max, permutations = P2._palp_PM_max(check=True)
        sage: PM_max
        [3 0 0]
        [0 3 0]
        [0 0 3]
        sage: list(permutations.values())
        [[(1,2,3), (1,2,3)],
         [(1,3,2), (1,3,2)],
         [(1,3), (1,3)],
         [(1,2), (1,2)],
         [(), ()],
         [(2,3), (2,3)]]
        sage: PM_max.automorphisms_of_rows_and_columns()
        [((), ()),
         ((1,2,3), (1,2,3)),
         ((1,3,2), (1,3,2)),
         ((2,3), (2,3)),
         ((1,2), (1,2)),
         ((1,3), (1,3))]
        sage: PMs = (i._palp_PM_max(check=True)
        ....:        for i in ReflexivePolytopes(2))
        sage: results = (len(i) == len(j.automorphisms_of_rows_and_columns())
        ....:            for j, i in PMs)
        sage: all(results)  # long time
        True

    TESTS:

    Check that a bug introduced in :issue:`35997` is fixed::

        sage: from sage.geometry.palp_normal_form import _palp_PM_max, _palp_canonical_order
        sage: P = Polyhedron([(-4,-6),(-4,-5),(0,0),(1,0),(5,6)])
        sage: PM = P.slack_matrix().transpose()
        sage: _palp_PM_max(PM)
        [9 5 4 0 0]
        [6 0 6 5 0]
        [1 5 0 0 4]
        [0 6 0 1 6]
        [0 0 3 5 3]
        sage: PM_max, permutations = _palp_PM_max(PM, check=True)
        sage: _palp_canonical_order(P.vertices(), PM_max, permutations)
        ([(1, 0), (0, 0), (2, 4), (1, 5), (-1, 1)], (1,2,3))
    """
    cdef int n_v = PM.ncols()
    cdef int n_f = PM.nrows()
    S_v = SymmetricGroup(n_v)
    S_f = SymmetricGroup(n_f)

    cdef int n_s = 1
    cdef dict permutations = {0: [S_f.one(), S_v.one()]}
    cdef int j, k, m, d
    cdef int element, max_element

    for j in range(n_v):
        max_element = PM.get_unsafe_int(0, (<PermutationGroupElement> permutations[0][1])(j + 1) - 1)
        m = 0
        for i in range(j + 1, n_v):
            element = PM.get_unsafe_int(0, (<PermutationGroupElement> permutations[0][1])(i + 1) - 1)
            if element > max_element:
                max_element = element
                m = i - j
        if m > 0:
            permutations[0][1] = (<PermutationGroupElement> permutations[0][1])._transpose_left(j + 1, m + j + 1)

    cdef first_row_index = 0

    # Arrange other rows one by one and compare with first row
    for k in range(1, n_f):
        # Error for k == 1 already!
        permutations[n_s] = [S_f.one(), S_v.one()]
        max_element = PM.get_unsafe_int(k, (<PermutationGroupElement> permutations[n_s][1])(1) - 1)
        m = 0
        for j in range(1, n_v):
            element = PM.get_unsafe_int(k, (<PermutationGroupElement> permutations[n_s][1])(j + 1) - 1)
            if element > max_element:
                max_element = element
                m = j
        if m > 0:
            permutations[n_s][1] = (<PermutationGroupElement> permutations[n_s][1])._transpose_left(1, m + 1)
        d = (PM.get_unsafe_int(k, (<PermutationGroupElement> permutations[n_s][1])(1) - 1)
             - PM.get_unsafe_int(first_row_index, (<PermutationGroupElement> permutations[0][1])(1) - 1))
        if d < 0:
            # The largest elt of this row is smaller than largest elt
            # in 1st row, so nothing to do
            continue
        # otherwise:
        for i in range(1, n_v):
            max_element = PM.get_unsafe_int(k, (<PermutationGroupElement> permutations[n_s][1])(i + 1) - 1)
            m = i
            for j in range(i + 1, n_v):
                element = PM.get_unsafe_int(k, (<PermutationGroupElement> permutations[n_s][1])(j + 1) - 1)
                if element > max_element:
                    max_element = element
                    m = j
            if m > i:
                permutations[n_s][1] = (<PermutationGroupElement> permutations[n_s][1])._transpose_left(i + 1, m + 1)
            if d == 0:
                d = (PM.get_unsafe_int(k, (<PermutationGroupElement> permutations[n_s][1])(i+1) - 1)
                     - PM.get_unsafe_int(first_row_index, (<PermutationGroupElement> permutations[0][1])(i + 1) - 1))
                if d < 0:
                    break
        if d < 0:
            # This row is smaller than 1st row, so nothing to do
            del permutations[n_s]
            continue
        permutations[n_s][0] = (<PermutationGroupElement> permutations[n_s][0])._transpose_left(1, k + 1)
        if d == 0:
            # This row is the same, so we have a symmetry!
            n_s += 1
        else:
            # This row is larger, so it becomes the first row and
            # the symmetries reset.
            first_row_index = k
            permutations = {0: permutations[n_s]}
            n_s = 1
    permutations = {k: permutations[k] for k in permutations if k < n_s}

    cdef tuple b = tuple(PM.get_unsafe_int((<PermutationGroupElement> permutations[0][0])(1) - 1,
                                           (<PermutationGroupElement> permutations[0][1])(j+1) - 1)
                         for j in range(n_v))
    # Work out the restrictions the current permutations
    # place on other permutations as a automorphisms
    # of the first row
    # The array is such that:
    # S = [i, 1, ..., 1 (ith), j, i+1, ..., i+1 (jth), k ... ]
    # describes the "symmetry blocks"
    cdef list S = list(range(1, n_v + 1))
    for i in range(1, n_v):
        if b[i-1] == b[i]:
            S[i] = S[i-1]
            S[S[i]-1] += 1
        else:
            S[i] = i + 1

    cdef int l, cf, ccf, n_s_bar, d1, v0, vc, vj
    cdef list l_r

    # We determine the other rows of PM_max in turn by use of perms and
    # aut on previous rows.
    for l in range(1, n_f - 1):
        n_s = len(permutations)
        n_s_bar = n_s
        cf = 0
        l_r = [0]*n_v
        # Search for possible local permutations based off previous
        # global permutations.
        for k in range(n_s_bar - 1, -1, -1):
            # number of local permutations associated with current global
            n_p = 0
            ccf = cf
            permutations_bar = {0: list(permutations[k])}
            # We look for the line with the maximal entry in the first
            # subsymmetry block, i.e. we are allowed to swap elements
            # between 0 and S(0)
            for s in range(l, n_f):
                for j in range(1, S[0]):
                    v0 = PM.get_unsafe_int((<PermutationGroupElement> permutations_bar[n_p][0])(s+1) - 1,
                                           (<PermutationGroupElement> permutations_bar[n_p][1])(1) - 1)
                    vj = PM.get_unsafe_int((<PermutationGroupElement> permutations_bar[n_p][0])(s+1) - 1,
                                           (<PermutationGroupElement> permutations_bar[n_p][1])(j+1) - 1)
                    if v0 < vj:
                        permutations_bar[n_p][1] = (<PermutationGroupElement> permutations_bar[n_p][1])._transpose_left(1, j + 1)
                if ccf == 0:
                    l_r[0] = PM.get_unsafe_int((<PermutationGroupElement> permutations_bar[n_p][0])(s+1) - 1,
                                               (<PermutationGroupElement> permutations_bar[n_p][1])(1) - 1)
                    if s != l:
                        permutations_bar[n_p][0] = (<PermutationGroupElement> permutations_bar[n_p][0])._transpose_left(l + 1, s + 1)
                    n_p += 1
                    ccf = 1
                    permutations_bar[n_p] = list(permutations[k])
                else:
                    d1 = PM.get_unsafe_int((<PermutationGroupElement> permutations_bar[n_p][0])(s+1) - 1,
                                           (<PermutationGroupElement> permutations_bar[n_p][1])(1) - 1)
                    d = d1 - l_r[0]
                    if d < 0:
                        # We move to the next line
                        continue
                    elif d==0:
                        # Maximal values agree, so possible symmetry
                        if s != l:
                            permutations_bar[n_p][0] = (<PermutationGroupElement> permutations_bar[n_p][0])._transpose_left(l + 1, s + 1)
                        n_p += 1
                        permutations_bar[n_p] = list(permutations[k])
                    else:
                        # We found a greater maximal value for first entry.
                        # It becomes our new reference:
                        l_r[0] = d1
                        if s != l:
                            permutations_bar[n_p][0] = (<PermutationGroupElement> permutations_bar[n_p][0])._transpose_left(l + 1, s + 1)
                        # Forget previous work done
                        cf = 0
                        permutations_bar = {0: list(permutations_bar[n_p])}
                        n_p = 1
                        permutations_bar[n_p] = list(permutations[k])
                        n_s = k + 1
            # Check if the permutations found just now work
            # with other elements
            for c in range(1, n_v):
                h = S[c]
                ccf = cf
                # Now let us find out where the end of the
                # next symmetry block is:
                if h < c + 1:
                    h = S[h - 1]
                s = n_p
                # Check through this block for each possible permutation
                while s > 0:
                    s -= 1
                    # Find the largest value in this symmetry block
                    for j in range(c + 1, h):
                        vc = PM.get_unsafe_int((<PermutationGroupElement> permutations_bar[s][0])(l+1) - 1,
                                               (<PermutationGroupElement> permutations_bar[s][1])(c+1) - 1)
                        vj = PM.get_unsafe_int((<PermutationGroupElement> permutations_bar[s][0])(l+1) - 1,
                                               (<PermutationGroupElement> permutations_bar[s][1])(j+1) - 1)
                        if vc < vj:
                            permutations_bar[s][1] = (<PermutationGroupElement> permutations_bar[s][1])._transpose_left(c + 1, j + 1)
                    if ccf == 0:
                        # Set reference and carry on to next permutation
                        l_r[c] = PM.get_unsafe_int((<PermutationGroupElement> permutations_bar[s][0])(l+1) - 1,
                                                   (<PermutationGroupElement> permutations_bar[s][1])(c+1) - 1)
                        ccf = 1
                    else:
                        d1 = PM.get_unsafe_int((<PermutationGroupElement> permutations_bar[s][0])(l+1) - 1,
                                               (<PermutationGroupElement> permutations_bar[s][1])(c+1) - 1)
                        d = d1 - l_r[c]
                        if d < 0:
                            n_p -= 1
                            if s < n_p:
                                permutations_bar[s] = list(permutations_bar[n_p])
                        elif d > 0:
                            # The current case leads to a smaller matrix,
                            # hence this case becomes our new reference
                            l_r[c] = d1
                            cf = 0
                            n_p = s + 1
                            n_s = k + 1
            # Update permutations
            if (n_s - 1) > k:
                permutations[k] = list(permutations[n_s - 1])
            n_s -= 1
            for s in range(n_p):
                permutations[n_s] = list(permutations_bar[s])
                n_s += 1
            cf = n_s
        permutations = {k: permutations[k] for k in permutations if k < n_s}
        # If the automorphisms are not already completely restricted,
        # update them
        if S != list(range(1, n_v + 1)):
            # Take the old automorphisms and update by
            # the restrictions the last worked out
            # row imposes.
            c = 0
            M = tuple(PM.get_unsafe_int((<PermutationGroupElement> permutations[0][0])(l+1) - 1,
                                        (<PermutationGroupElement> permutations[0][1])(j+1) - 1)
                      for j in range(n_v))
            while c < n_v:
                s = S[c] + 1
                S[c] = c + 1
                c += 1
                while c < (s - 1):
                    if M[c] == M[c - 1]:
                        S[c] = S[c - 1]
                        S[S[c] - 1] += 1
                    else:
                        S[c] = c + 1
                    c += 1
    # Now we have the perms, we construct PM_max using one of them
    PM_max = PM.with_permuted_rows_and_columns(*permutations[0])
    if check:
        return (PM_max, permutations)
    else:
        return PM_max


def _palp_canonical_order(vertices, PM_max, permutations):
    r"""
    Compute the PALP normal form of vertices of a lattice polytope
    using auxiliary data computed elsewhere.

    This is a helper function for
    :meth:`~sage.geometry.lattice_polytope.LatticePolytopeClass.normal_form`
    and should not be called directly.

    Given an iterable of vertices, the maximal vertex-facet pairing matrix
    and the permutations realizing this matrix, apply the last part of the
    PALP algorithm and return the normal form.

    INPUT:

    - ``vertices`` -- iterable of iterables; the vertices

    - ``PM_max`` -- the maximal vertex-facet pairing matrix

    - ``permutation`` -- the permutations of the vertices yielding ``PM_max``

    OUTPUT: the PALP normal form as an iterable of integer vectors

    TESTS::

        sage: L = lattice_polytope.cross_polytope(2)
        sage: V = L.vertices()
        sage: PM_max, permutations = L._palp_PM_max(check=True)
        sage: from sage.geometry.lattice_polytope import _palp_canonical_order
        sage: _palp_canonical_order(V, PM_max, permutations)
        (M( 1,  0),
         M( 0,  1),
         M( 0, -1),
         M(-1,  0)
         in 2-d lattice M, (1,3,2,4))
    """
    n_v = PM_max.ncols()
    S_v = SymmetricGroup(n_v)
    p_c = S_v.one()
    M_max = [max(row[j] for row in PM_max.rows()) for j in range(n_v)]
    S_max = sum(PM_max)
    for i in range(n_v):
        k = i
        for j in range(i + 1, n_v):
            if M_max[j] < M_max[k] or \
               (M_max[j] == M_max[k] and S_max[j] < S_max[k]):
                k = j
        if not k == i:
            M_max[i], M_max[k] = M_max[k], M_max[i]
            S_max[i], S_max[k] = S_max[k], S_max[i]
            p_c = p_c._transpose_left(1 + i, 1 + k)
    # Create array of possible NFs.
    permutations = [p_c * l[1] for l in permutations.values()]
    if isinstance(vertices, Matrix):
        Vmatrix = vertices
    else:
        Vmatrix = column_matrix(vertices)
    Vs = [(Vmatrix.with_permuted_columns(sig).hermite_form(), sig)
          for sig in permutations]
    Vmin = min(Vs, key=lambda x: x[0])
    return Vmin[0].columns(), Vmin[1]



# sage_setup: distribution = sagemath-groups
"""
Artin Groups

Artin groups are implemented as a particular case of finitely presented
groups. For finite-type Artin groups, there is a specific left normal
form using the Garside structure associated to the lift the long element
of the corresponding Coxeter group.

AUTHORS:

- Travis Scrimshaw (2018-02-05): Initial version
- Travis Scrimshaw (2025-01-30): Allowed general construction; implemented
  general Burau representation

.. TODO::

    Implement affine type Artin groups with their well-known embeddings into
    the classical braid group.
"""

# ****************************************************************************
#       Copyright (C) 2018-2025 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
# *****************************************************************************

from sage.combinat.root_system.coxeter_matrix import CoxeterMatrix
from sage.combinat.root_system.coxeter_group import CoxeterGroup
from sage.groups.free_group import FreeGroup
from sage.groups.finitely_presented import FinitelyPresentedGroup, FinitelyPresentedGroupElement
from sage.misc.cachefunc import cached_method
from sage.misc.lazy_attribute import lazy_attribute
from sage.rings.infinity import Infinity
from sage.structure.richcmp import richcmp, rich_to_bool
from sage.structure.unique_representation import UniqueRepresentation


class ArtinGroupElement(FinitelyPresentedGroupElement):
    """
    An element of an Artin group.

    It is a particular case of element of a finitely presented group.

    EXAMPLES::

        sage: # needs sage.rings.number_field
        sage: A.<s1,s2,s3> = ArtinGroup(['B',3])
        sage: A
        Artin group of type ['B', 3]
        sage: s1 * s2 / s3 / s2
        s1*s2*s3^-1*s2^-1
        sage: A((1, 2, -3, -2))
        s1*s2*s3^-1*s2^-1
    """
    def _latex_(self):
        r"""
        Return a LaTeX representation of ``self``.

        OUTPUT: string; a valid LaTeX math command sequence

        TESTS::

            sage: A = ArtinGroup(['B', 3])                                              # needs sage.rings.number_field
            sage: b = A([1, 2, 3, -1, 2, -3])                                           # needs sage.rings.number_field
            sage: b._latex_()                                                           # needs sage.rings.number_field
            '\\sigma_{1}\\sigma_{2}\\sigma_{3}\\sigma_{1}^{-1}\\sigma_{2}\\sigma_{3}^{-1}'

            sage: # needs sage.combinat sage.graphs
            sage: B = BraidGroup(4)
            sage: b = B([1, 2, 3, -1, 2, -3])
            sage: b._latex_()
            '\\sigma_{1}\\sigma_{2}\\sigma_{3}\\sigma_{1}^{-1}\\sigma_{2}\\sigma_{3}^{-1}'
            sage: B.one()._latex_()
            '1'
        """
        word = self.Tietze()
        if not word:
            return '1'
        return ''.join(r"\sigma_{%s}^{-1}" % (-i) if i < 0 else r"\sigma_{%s}" % i
                       for i in word)

    def exponent_sum(self):
        """
        Return the exponent sum of ``self``.

        OUTPUT: integer

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: A = ArtinGroup(['E',6])
            sage: b = A([1, 4, -3, 2])
            sage: b.exponent_sum()
            2
            sage: b = A([])
            sage: b.exponent_sum()
            0

            sage: # needs sage.combinat sage.graphs
            sage: B = BraidGroup(5)
            sage: b = B([1, 4, -3, 2])
            sage: b.exponent_sum()
            2
            sage: b = B([])
            sage: b.exponent_sum()
            0
        """
        return sum(s.sign() for s in self.Tietze())

    def coxeter_group_element(self, W=None):
        """
        Return the corresponding Coxeter group element under the natural
        projection.

        INPUT:

        - ``W`` -- (default: ``self.parent().coxeter_group()``) the image
          Coxeter group

        OUTPUT: an element of the Coxeter group  ``W``

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: B.<s1,s2,s3> = ArtinGroup(['B',3])
            sage: b = s1 * s2 / s3 / s2
            sage: b1 = b.coxeter_group_element(); b1
            [ 1 -1  0]
            [ 2 -1  0]
            [ a -a  1]
            sage: b.coxeter_group_element().reduced_word()
            [1, 2, 3, 2]

            sage: # needs sage.combinat sage.rings.number_field
            sage: A.<s1,s2,s3> = ArtinGroup(['A',3])
            sage: c = s1 * s2 *s3
            sage: c1 = c.coxeter_group_element(); c1
            [4, 1, 2, 3]
            sage: c1.reduced_word()
            [3, 2, 1]
            sage: c.coxeter_group_element(W=SymmetricGroup(4))
            (1,4,3,2)

            sage: # needs sage.graphs
            sage: A.<s1,s2,s3> = BraidGroup(4)
            sage: c = s1 * s2 * s3^-1
            sage: c0 = c.coxeter_group_element(); c0
            [4, 1, 2, 3]
            sage: c1 = c.coxeter_group_element(W=SymmetricGroup(4)); c1
            (1,4,3,2)

        From an element of the Coxeter group it is possible to recover
        the image by the standard section to the Artin group::

            sage: # needs sage.rings.number_field
            sage: B(b1)
            s1*s2*s3*s2

            sage: # needs sage.graphs
            sage: A(c0)
            s1*s2*s3
            sage: A(c0) == A(c1)
            True
        """
        if W is None:
            W = self.parent().coxeter_group()
        s = W.simple_reflections()
        In = W.index_set()
        return W.prod(s[In[abs(i) - 1]] for i in self.Tietze())

    def burau_matrix(self, var='t'):
        r"""
        Return the Burau matrix of the Artin group element.

        Following [BQ2024]_, the (generalized) Burau representation
        of an Artin group is defined by deforming the reflection
        representation of the corresponding Coxeter group. However,
        we substitute `q \mapsto -t` from [BQ2024]_ to match one of
        the unitary (reduced) Burau representations of the braid group
        (see :meth:`sage.groups.braid.Braid.burau_matrix()` for details.)

        More precisely, let `(m_{ij})_{i,j \in I}` be the
        :meth:`Coxeter matrix<coxeter_matrix>`. Then the action is
        given on the basis `(\alpha_1, \ldots \alpha_n)` (corresponding
        to the reflection representation of the corresponding
        :meth:`Coxeter group<coxeter_group>`) by

        .. MATH::

            \sigma_i(\alpha_j) = \alpha_j
              - \langle \alpha_i, \alpha_j \rangle_q \alpha_i,
            \qquad \text{ where }
            \langle \alpha_i, \alpha_j \rangle_q := \begin{cases}
             1 + t^2 & \text{if } i = j, \\
             -2 t \cos(\pi/m_{ij}) & \text{if } i \neq j.
            \end{cases}.

        By convention `\cos(\pi/\infty) = 1`. Note that the inverse of the
        generators act by `\sigma_i^{-1}(\alpha_j) = \alpha_j - q^{-2}
        \langle \alpha_j, \alpha_i \rangle_q \alpha_i`.

        INPUT:

        - ``var`` -- string (default: ``'t'``); the name of the
          variable in the entries of the matrix

        OUTPUT:

        The Burau matrix of the Artin group element over the Laurent
        polynomial ring in the variable ``var``.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: A.<s1,s2,s3> = ArtinGroup(['B',3])
            sage: B1 = s1.burau_matrix()
            sage: B2 = s2.burau_matrix()
            sage: B3 = s3.burau_matrix()
            sage: [B1, B2, B3]
            [
            [-t^2    t    0]  [   1    0    0]  [   1    0    0]
            [   0    1    0]  [   t -t^2  a*t]  [   0    1    0]
            [   0    0    1], [   0    0    1], [   0  a*t -t^2]
            ]
            sage: B1 * B2 * B1 == B2 * B1 * B2
            True
            sage: B2 * B3 * B2 * B3 == B3 * B2 * B3 * B2
            True
            sage: B1 * B3 == B3 * B1
            True
            sage: (~s1).burau_matrix() * B1 == 1
            True

        We verify the example in Theorem 4.1 of [BQ2024]_::

            sage: # needs sage.graphs
            sage: A.<s1,s2,s3,s4> = ArtinGroup(['A', 3, 1])
            sage: [g.burau_matrix() for g in A.gens()]
            [
            [-t^2    t    0    t]  [   1    0    0    0]  [   1    0    0    0]
            [   0    1    0    0]  [   t -t^2    t    0]  [   0    1    0    0]
            [   0    0    1    0]  [   0    0    1    0]  [   0    t -t^2    t]
            [   0    0    0    1], [   0    0    0    1], [   0    0    0    1],
            <BLANKLINE>
            [   1    0    0    0]
            [   0    1    0    0]
            [   0    0    1    0]
            [   t    0    t -t^2]
            ]
            sage: a = s3^2 * s4 * s3 * s2 *s1 * ~s3 * s4 * s3 * s2 * s1^-2 * s4
            sage: b = s1^2 * ~s2 * s4 * s1 * ~s3 * s2 * ~s4 * s3 * s1 * s4 * s1 * ~s2 * s4^-2 * s3
            sage: alpha = a * s3 * ~a
            sage: beta = b * s2 * ~b
            sage: elm = alpha * beta * ~alpha * ~beta
            sage: print(elm.Tietze())
            (3, 3, 4, 3, 2, 1, -3, 4, 3, 2, -1, -1, 4, 3, -4, 1, 1, -2, -3,
             -4, 3, -1, -2, -3, -4, -3, -3, 1, 1, -2, 4, 1, -3, 2, -4, 3, 1,
             4, 1, -2, -4, -4, 3, 2, -3, 4, 4, 2, -1, -4, -1, -3, 4, -2, 3,
             -1, -4, 2, -1, -1, 3, 3, 4, 3, 2, 1, -3, 4, 3, 2, -1, -1, 4,
             -3, -4, 1, 1, -2, -3, -4, 3, -1, -2, -3, -4, -3, -3, 1, 1, -2,
             4, 1, -3, 2, -4, 3, 1, 4, 1, -2, -4, -4, 3, -2, -3, 4, 4, 2,
             -1, -4, -1, -3, 4, -2, 3, -1, -4, 2, -1, -1)
            sage: elm.burau_matrix()
            [1 0 0 0]
            [0 1 0 0]
            [0 0 1 0]
            [0 0 0 1]

        Next, we show ``elm`` is not the identity by using the embedding of
        the affine braid group `\widetilde{B}_n \to B_{n+1}`::

            sage: # needs sage.libs.braiding
            sage: B.<t1,t2,t3,t4> = BraidGroup(5)
            sage: D = t1 * t2 * t3 * t4^2
            sage: t0 = D * t3 * ~D
            sage: t0*t1*t0 == t1*t0*t1
            True
            sage: t0*t2 == t2*t0
            True
            sage: t0*t3*t0 == t3*t0*t3
            True
            sage: T = [t0, t1, t2, t3]
            sage: emb = B.prod(T[i-1] if i > 0 else ~T[-i-1] for i in elm.Tietze())
            sage: emb.is_one()
            False

        Since the Burau representation does not respect the group embedding,
        the corresponding `B_5` element's Burau matrix is not the identity
        (Bigelow gave an example of the representation not being faithful for
        `B_5`, but it is still open for `B_4`)::

            sage: emb.burau_matrix() != 1                                               # needs sage.libs.braiding
            True

        We also verify the result using the elements in [BQ2024]_ Remark 4.2::

            sage: ap = s3 * s1 * s2 * s1 * ~s3 * s4 * s2 * s3 * s2 * ~s3 * s1^-2 * s4
            sage: bp = s1 * ~s4 * s1^2 * s3^-2 * ~s2 * s4 * s1 * ~s3 * s2 * ~s4 * s3 * s1 * s4 * s1 * ~s2 * s4^-2 * s3
            sage: alpha = ap * s3 * ~ap
            sage: beta = bp * s2 * ~bp
            sage: elm = alpha * beta * ~alpha * ~beta                                   # needs sage.libs.braiding
            sage: elm.burau_matrix()                                                    # needs sage.libs.braiding
            [1 0 0 0]
            [0 1 0 0]
            [0 0 1 0]
            [0 0 0 1]

        REFERENCES:

        - [BQ2024]_
        """
        gens, invs = self.parent()._burau_generators
        MS = gens[0].parent()
        ret = MS.prod(gens[i-1] if i > 0 else invs[-i-1] for i in self.Tietze())

        if var == 't':
            return ret

        from sage.rings.polynomial.laurent_polynomial_ring import LaurentPolynomialRing
        poly_ring = LaurentPolynomialRing(ret.base_ring().base_ring(), var)
        return ret.change_ring(poly_ring)


class FiniteTypeArtinGroupElement(ArtinGroupElement):
    """
    An element of a finite-type Artin group.
    """
    def _richcmp_(self, other, op):
        """
        Compare ``self`` and ``other``.

        TESTS::

            sage: # needs sage.rings.number_field
            sage: A = ArtinGroup(['B',3])
            sage: x = A([1, 2, 1])
            sage: y = A([2, 1, 2])
            sage: x == y
            True
            sage: x < y^(-1)
            True
            sage: A([]) == A.one()
            True
            sage: x = A([2, 3, 2, 3])
            sage: y = A([3, 2, 3, 2])
            sage: x == y
            True
            sage: x < y^(-1)
            True
        """
        if self.Tietze() == other.Tietze():
            return rich_to_bool(op, 0)
        nfself = [i.Tietze() for i in self.left_normal_form()]
        nfother = [i.Tietze() for i in other.left_normal_form()]
        return richcmp(nfself, nfother, op)

    def __hash__(self):
        r"""
        Return a hash value for ``self``.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: B.<s1,s2,s3> = ArtinGroup(['B',3])
            sage: hash(s1*s3) == hash(s3*s1)
            True
            sage: hash(s1*s2) == hash(s2*s1)
            False
            sage: hash(s1*s2*s1) == hash(s2*s1*s2)
            True
            sage: hash(s2*s3*s2) == hash(s3*s2*s3)
            False
            sage: hash(s2*s3*s2*s3) == hash(s3*s2*s3*s2)
            True
        """
        return hash(tuple(i.Tietze() for i in self.left_normal_form()))

    @cached_method
    def left_normal_form(self):
        r"""
        Return the left normal form of ``self``.

        OUTPUT:

        A tuple of simple generators in the left normal form. The first
        element is a power of `\Delta`, and the rest are elements of the
        natural section lift from the corresponding Coxeter group.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: A = ArtinGroup(['B',3])
            sage: A([1]).left_normal_form()
            (1, s1)
            sage: A([-1]).left_normal_form()
            (s1^-1*(s2^-1*s1^-1*s3^-1)^2*s2^-1*s3^-1, s3*(s2*s3*s1)^2*s2)
            sage: A([1, 2, 2, 1, 2]).left_normal_form()
            (1, s1*s2*s1, s2*s1)
            sage: A([3, 3, -2]).left_normal_form()
            (s1^-1*(s2^-1*s1^-1*s3^-1)^2*s2^-1*s3^-1,
             s3*s1*s2*s3*s2*s1, s3, s3*s2*s3)
            sage: A([1, 2, 3, -1, 2, -3]).left_normal_form()
            (s1^-1*(s2^-1*s1^-1*s3^-1)^2*s2^-1*s3^-1,
             (s3*s1*s2)^2*s1, s1*s2*s3*s2)
            sage: A([1,2,1,3,2,1,3,2,3,3,2,3,1,2,3,1,2,3,1,2]).left_normal_form()
            ((s3*(s2*s3*s1)^2*s2*s1)^2, s3*s2)

            sage: # needs sage.combinat
            sage: B = BraidGroup(4)
            sage: b = B([1, 2, 3, -1, 2, -3])
            sage: b.left_normal_form()                                                  # needs sage.libs.braiding
            (s0^-1*s1^-1*s0^-1*s2^-1*s1^-1*s0^-1, s0*s1*s2*s1*s0, s0*s2*s1)
            sage: c = B([1])
            sage: c.left_normal_form()                                                  # needs sage.libs.braiding
            (1, s0)
        """
        lnfp = self._left_normal_form_coxeter()
        P = self.parent()
        return tuple([P.delta() ** lnfp[0]] +
                     [P._standard_lift(w) for w in lnfp[1:]])

    def _left_normal_form_coxeter(self):
        r"""
        Return the left normal form of the element, in the `\Delta`
        exponent and Coxeter group element form.

        OUTPUT: tuple whose first element is the power of `\Delta`, and the
        rest are the Coxeter elements corresponding to the simple factors

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: A = ArtinGroup(['E',6])
            sage: A([2, -4, 2, 3, 1, 3, 2, 1, -2])._left_normal_form_coxeter()
            (
                [ 0  0  0  0  0 -1]  [ 0  0 -1  1  0  0]  [-1  0  1  0  0  0]
                [ 0  1  0 -1  0  0]  [ 0 -1  0  1  0  0]  [ 0 -1  0  1  0  0]
                [ 0  0  0  0 -1  0]  [-1  0  0  1  0  0]  [ 0  0  1  0  0  0]
                [ 0  1 -1  0 -1  0]  [-1 -1  0  1  1  0]  [ 0  0  0  1  0  0]
                [ 0  0 -1  0  0  0]  [ 0  0  0  0  1  0]  [ 0  0  0  0  1  0]
            -1, [-1  0  0  0  0  0], [ 0  0  0  0  0  1], [ 0  0  0  0  0  1]
            )
            sage: A = ArtinGroup(['F',4])
            sage: A([2, 3, -4, 2, 3, -2, 1, -2, 3, 4, 1, -2])._left_normal_form_coxeter()
            (
                [-1  0  0  0]  [ -1   1   0   0]  [-1  0  0  0]
                [-1 -1  a -a]  [ -2   3  -a   0]  [-1  1 -a  0]
                [-a  0  1 -2]  [ -a 2*a  -1  -1]  [ 0  0 -1  0]
            -3, [-a  0  1 -1], [ -a   a   0  -1], [ 0  0  0 -1],
            <BLANKLINE>
            [  -1    1    0   -a]  [  -1    0    a   -a]
            [   0    1    0 -2*a]  [  -1    1    a -2*a]
            [   0    a   -1   -2]  [   0    0    2   -3]
            [   0    a   -1   -1], [   0    0    1   -1]
            )
        """
        delta = 0
        Delta = self.parent().coxeter_group().long_element()
        sr = self.parent().coxeter_group().simple_reflections()
        tz = self.Tietze()
        if tz == ():
            return (0,)
        form = []
        for i in tz:
            if i > 0:
                form.append(sr[i])
            else:
                delta += 1
                form = [Delta * a * Delta for a in form]
                form.append(Delta * sr[-i])
        i = j = 0
        while j < len(form):
            while i < len(form) - j - 1:
                e = form[i].descents(side='right')
                s = form[i + 1].descents(side='left')
                S = set(s).difference(set(e))
                while S:
                    a = list(S)[0]
                    form[i] = form[i] * sr[a]
                    form[i + 1] = sr[a] * form[i+1]
                    e = form[i].descents(side='right')
                    s = form[i + 1].descents(side='left')
                    S = set(s).difference(set(e))
                if form[i+1].length() == 0:
                    form.pop(i+1)
                    i = 0
                else:
                    i += 1
            j += 1
            i = 0
        form = [elt for elt in form if elt.length()]
        while form and form[0] == Delta:
            form.pop(0)
            delta -= 1
        return tuple([-delta] + form)


class ArtinGroup(UniqueRepresentation, FinitelyPresentedGroup):
    r"""
    An Artin group.

    Fix an index set `I`. Let `M = (m_{ij})_{i,j \in I}` be a
    :class:`Coxeter matrix
    <sage.combinat.root_system.coxeter_matrix.CoxeterMatrix>`.
    An *Artin group* is a group `A_M` that has a presentation
    given by generators `\{ s_i \mid i \in I \}` and relations

    .. MATH::

        \underbrace{s_i s_j s_i \cdots}_{m_{ij}}
        = \underbrace{s_j s_i s_j \cdots}_{\text{$m_{ji}$ factors}}

    for all `i,j \in I` with the usual convention that `m_{ij} = \infty`
    implies no relation between `s_i` and `s_j`. There is a natural
    corresponding Coxeter group `W_M` by imposing the additional
    relations `s_i^2 = 1` for all `i \in I`. Furthermore, there is
    a natural section of `W_M` by sending a reduced word
    `s_{i_1} \cdots s_{i_{\ell}} \mapsto s_{i_1} \cdots s_{i_{\ell}}`.

    Artin groups `A_M` are classified based on the Coxeter data:

    - `A_M` is of *finite type* or *spherical* if `W_M` is finite;
    - `A_M` is of *affine type* if `W_M` is of affine type;
    - `A_M` is of *large type* if `m_{ij} \geq 4` for all `i,j \in I`;
    - `A_M` is of *extra-large type* if `m_{ij} \geq 5` for all `i,j \in I`;
    - `A_M` is *right-angled* if `m_{ij} \in \{2,\infty\}` for all `i,j \in I`.

    Artin groups are conjectured to have many nice properties:

    - Artin groups are torsion free.
    - Finite type Artin groups have `Z(A_M) = \ZZ` and infinite type
      Artin groups have trivial center.
    - Artin groups have solvable word problems.
    - `H_{W_M} / W_M` is a `K(A_M, 1)`-space, where `H_W` is the
      hyperplane complement of the Coxeter group `W` acting on `\CC^n`.

    These conjectures are known when the Artin group is finite type and a
    number of other cases. See, e.g., [GP2012]_ and references therein.

    INPUT:

    - ``coxeter_data`` -- data defining a Coxeter matrix

    - ``names`` -- string or list/tuple/iterable of strings
      (default: ``'s'``); the generator names or name prefix

    EXAMPLES::

        sage: A.<a,b,c> = ArtinGroup(['B',3]);  A                                       # needs sage.rings.number_field
        Artin group of type ['B', 3]
        sage: ArtinGroup(['B',3])                                                       # needs sage.rings.number_field
        Artin group of type ['B', 3]

    The input must always include the Coxeter data, but the ``names``
    can either be a string representing the prefix of the names or
    the explicit names of the generators. Otherwise the default prefix
    of ``'s'`` is used::

        sage: ArtinGroup(['B',2]).generators()                                          # needs sage.rings.number_field
        (s1, s2)
        sage: ArtinGroup(['B',2], 'g').generators()                                     # needs sage.rings.number_field
        (g1, g2)
        sage: ArtinGroup(['B',2], 'x,y').generators()                                   # needs sage.rings.number_field
        (x, y)

    REFERENCES:

    - :wikipedia:`Artin_group`

    .. SEEALSO::

        :class:`~sage.groups.raag.RightAngledArtinGroup`
    """
    @staticmethod
    def __classcall_private__(cls, coxeter_data, names=None):
        """
        Normalize input to ensure a unique representation.

        TESTS::

            sage: # needs sage.rings.number_field
            sage: A1 = ArtinGroup(['B',3])
            sage: A2 = ArtinGroup(['B',3], 's')
            sage: A3 = ArtinGroup(['B',3],  ['s1','s2','s3'])
            sage: A1 is A2 and A2 is A3
            True

            sage: # needs sage.rings.number_field
            sage: A1 = ArtinGroup(['B',2], 'a,b')
            sage: A2 = ArtinGroup([[1,4],[4,1]], 'a,b')
            sage: A3.<a,b> = ArtinGroup('B2')
            sage: A1 is A2 and A2 is A3
            True

            sage: ArtinGroup(['A',3]) is BraidGroup(4, 's1,s2,s3')                      # needs sage.combinat sage.rings.number_field
            True

            sage: # needs sage.graphs sage.rings.number_field
            sage: G = graphs.PathGraph(3)
            sage: CM = CoxeterMatrix([[1,-1,2],[-1,1,-1],[2,-1,1]], index_set=G.vertices(sort=True))
            sage: A = groups.misc.Artin(CM)
            sage: Ap = groups.misc.RightAngledArtin(G, 's')
            sage: A is Ap
            True
        """
        coxeter_data = CoxeterMatrix(coxeter_data)
        if names is None:
            names = 's'
        if isinstance(names, str):
            if ',' in names:
                names = [x.strip() for x in names.split(',')]
            else:
                names = [names + str(i) for i in coxeter_data.index_set()]
        names = tuple(names)
        if len(names) != coxeter_data.rank():
            raise ValueError("the number of generators must match"
                             " the rank of the Coxeter type")
        if all(m == Infinity for m in coxeter_data.coxeter_graph().edge_labels()):
            from sage.groups.raag import RightAngledArtinGroup
            return RightAngledArtinGroup(coxeter_data.coxeter_graph(), names)
        if not coxeter_data.is_finite():
            return super().__classcall__(cls, coxeter_data, names)
        if coxeter_data.coxeter_type().cartan_type().type() == 'A':
            from sage.groups.braid import BraidGroup
            return BraidGroup(coxeter_data.rank() + 1, names)
        return FiniteTypeArtinGroup(coxeter_data, names)

    def __init__(self, coxeter_matrix, names) -> None:
        """
        Initialize ``self``.

        TESTS::

            sage: # needs sage.rings.number_field
            sage: A = ArtinGroup(['D',4])
            sage: TestSuite(A).run()
            sage: A = ArtinGroup(['B',3], ['x','y','z'])
            sage: TestSuite(A).run()
        """
        self._coxeter_group = CoxeterGroup(coxeter_matrix)
        free_group = FreeGroup(names)
        rels = []
        # Generate the relations based on the Coxeter graph
        I = coxeter_matrix.index_set()
        gens = free_group.gens()
        for ii, i in enumerate(I):
            for jj, j in enumerate(I[ii + 1:], start=ii + 1):
                m = coxeter_matrix[i, j]
                if m == Infinity:  # no relation
                    continue
                elt = (gens[ii] * gens[jj]) ** (m // 2)
                if m % 2 == 1:
                    elt = elt * gens[ii] * ~gens[jj]
                elt = elt * (~gens[ii] * ~gens[jj]) ** (m // 2)
                rels.append(elt)
        FinitelyPresentedGroup.__init__(self, free_group, tuple(rels))

    def _repr_(self) -> str:
        """
        Return a string representation of ``self``.

        TESTS::

            sage: ArtinGroup(['B',3])                                                   # needs sage.rings.number_field
            Artin group of type ['B', 3]
            sage: ArtinGroup(['D',4], 'g')                                              # needs sage.rings.number_field
            Artin group of type ['D', 4]
        """
        try:
            data = self.coxeter_type().cartan_type()
            return "Artin group of type {}".format(data)
        except AttributeError:
            pass
        return "Artin group with Coxeter matrix:\n{}".format(self.coxeter_matrix())

    def cardinality(self):
        """
        Return the number of elements of ``self``.

        OUTPUT: infinity

        EXAMPLES::

            sage: # needs sage.graphs
            sage: Gamma = graphs.CycleGraph(5)
            sage: G = RightAngledArtinGroup(Gamma)
            sage: G.cardinality()
            +Infinity

            sage: A = ArtinGroup(['A',1])                                               # needs sage.rings.number_field
            sage: A.cardinality()                                                       # needs sage.rings.number_field
            +Infinity
        """
        from sage.rings.infinity import Infinity
        return Infinity

    order = cardinality

    def as_permutation_group(self):
        """
        Return an isomorphic permutation group.

        This raises a :exc:`ValueError` error since Artin groups are
        infinite and have no corresponding permutation group.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: Gamma = graphs.CycleGraph(5)
            sage: G = RightAngledArtinGroup(Gamma)
            sage: G.as_permutation_group()
            Traceback (most recent call last):
            ...
            ValueError: the group is infinite

            sage: A = ArtinGroup(['D',4], 'g')                                          # needs sage.rings.number_field
            sage: A.as_permutation_group()                                              # needs sage.rings.number_field
            Traceback (most recent call last):
            ...
            ValueError: the group is infinite
        """
        raise ValueError("the group is infinite")

    def coxeter_type(self):
        """
        Return the Coxeter type of ``self``.

        EXAMPLES::

            sage: A = ArtinGroup(['D',4])                                               # needs sage.rings.number_field
            sage: A.coxeter_type()                                                      # needs sage.rings.number_field
            Coxeter type of ['D', 4]
        """
        return self._coxeter_group.coxeter_type()

    def coxeter_matrix(self):
        """
        Return the Coxeter matrix of ``self``.

        EXAMPLES::

            sage: A = ArtinGroup(['B',3])                                               # needs sage.rings.number_field
            sage: A.coxeter_matrix()                                                    # needs sage.rings.number_field
            [1 3 2]
            [3 1 4]
            [2 4 1]
        """
        return self._coxeter_group.coxeter_matrix()

    def coxeter_group(self):
        """
        Return the Coxeter group of ``self``.

        EXAMPLES::

            sage: A = ArtinGroup(['D',4])                                               # needs sage.rings.number_field
            sage: A.coxeter_group()                                                     # needs sage.rings.number_field
            Finite Coxeter group over Integer Ring with Coxeter matrix:
            [1 3 2 2]
            [3 1 3 3]
            [2 3 1 2]
            [2 3 2 1]
        """
        return self._coxeter_group

    def index_set(self):
        """
        Return the index set of ``self``.

        OUTPUT: tuple

        EXAMPLES::

            sage: A = ArtinGroup(['E',7])                                               # needs sage.rings.number_field
            sage: A.index_set()                                                         # needs sage.rings.number_field
            (1, 2, 3, 4, 5, 6, 7)
        """
        return self._coxeter_group.index_set()

    def _element_constructor_(self, x):
        """
        TESTS::

            sage: A = ArtinGroup(['B',3])                                               # needs sage.rings.number_field
            sage: A([2,1,-2,3,3,3,1])                                                   # needs sage.rings.number_field
            s2*s1*s2^-1*s3^3*s1
        """
        if x in self._coxeter_group:
            return self._standard_lift(x)
        return self.element_class(self, x)

    @cached_method
    def _an_element_(self):
        """
        Return an element of ``self``.

        EXAMPLES::

            sage: A = ArtinGroup(['B',2])                                               # needs sage.rings.number_field
            sage: A.an_element()                                                        # needs sage.rings.number_field
            s1
        """
        return self.gen(0)

    def some_elements(self) -> list:
        """
        Return a list of some elements of ``self``.

        EXAMPLES::

            sage: A = ArtinGroup(['B',3])                                               # needs sage.rings.number_field
            sage: A.some_elements()                                                     # needs sage.rings.number_field
            [s1, s1*s2*s3, (s1*s2*s3)^3]
        """
        rank = self.coxeter_matrix().rank()
        elements_list = [self.gen(0)]
        elements_list.append(self.prod(self.gens()))
        elements_list.append(elements_list[-1] ** min(rank, 3))
        return elements_list

    def _standard_lift_Tietze(self, w):
        """
        Return a Tietze word representing the Coxeter element ``w``
        under the natural section.

        INPUT:

        - ``w`` -- an element of the Coxeter group of ``self``

        EXAMPLES::

            sage: A = ArtinGroup(['B',3])                                               # needs sage.rings.number_field
            sage: A._standard_lift_Tietze(A.coxeter_group().long_element())             # needs sage.rings.number_field
            [3, 2, 3, 1, 2, 3, 1, 2, 1]
        """
        return w.reduced_word()

    @cached_method
    def _standard_lift(self, w):
        """
        Return the element of ``self`` that corresponds to the given
        Coxeter element ``w`` under the natural section.

        INPUT:

        - ``w`` -- an element of the Coxeter group of ``self``

        EXAMPLES::

            sage: A = ArtinGroup(['B',3])                                               # needs sage.combinat sage.rings.number_field
            sage: A._standard_lift(A.coxeter_group().long_element())                    # needs sage.combinat sage.rings.number_field
            s3*(s2*s3*s1)^2*s2*s1

            sage: B = BraidGroup(5)
            sage: P = Permutation([5, 3, 1, 2, 4])
            sage: B._standard_lift(P)
            s0*s1*s2*s3*s0*s1
        """
        return self(self._standard_lift_Tietze(w))

    @lazy_attribute
    def _burau_generators(self):
        """
        The Burau matrices for the generators of ``self`` and their inverses.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: A = ArtinGroup(['G',2])
            sage: A._burau_generators
            [[
            [-t^2  a*t]  [   1    0]
            [   0    1], [ a*t -t^2]
            ],
             [
            [ -t^-2 a*t^-1]  [     1      0]
            [     0      1], [a*t^-1  -t^-2]
            ]]
            sage: A = ArtinGroup(['H',3])
            sage: A._burau_generators
            [[
            [-t^2    t    0]  [              1               0               0]
            [   0    1    0]  [              t            -t^2 (1/2*a + 1/2)*t]
            [   0    0    1], [              0               0               1],
            <BLANKLINE>
            [              1               0               0]
            [              0               1               0]
            [              0 (1/2*a + 1/2)*t            -t^2]
            ],
             [
            [-t^-2  t^-1     0]
            [    0     1     0]
            [    0     0     1],
            <BLANKLINE>
            [                 1                  0                  0]
            [              t^-1              -t^-2 (1/2*a + 1/2)*t^-1]
            [                 0                  0                  1],
            <BLANKLINE>
            [                 1                  0                  0]
            [                 0                  1                  0]
            [                 0 (1/2*a + 1/2)*t^-1              -t^-2]
            ]]
            sage: CM = matrix([[1,4,7], [4,1,10], [7,10,1]])
            sage: A = ArtinGroup(CM)
            sage: gens = A._burau_generators[0]; gens
            [
            [                -t^2    (E(8) - E(8)^3)*t (-E(7)^3 - E(7)^4)*t]
            [                   0                    1                    0]
            [                   0                    0                    1],
            <BLANKLINE>
            [                  1                   0                   0]
            [  (E(8) - E(8)^3)*t                -t^2 (E(20) - E(20)^9)*t]
            [                  0                   0                   1],
            <BLANKLINE>
            [                   1                    0                    0]
            [                   0                    1                    0]
            [(-E(7)^3 - E(7)^4)*t  (E(20) - E(20)^9)*t                 -t^2]
            ]
            sage: all(gens[i] * A._burau_generators[1][i] == 1 for i in range(3))
            True
            sage: B1, B2, B3 = gens
            sage: (B1 * B2)^2 == (B2 * B1)^2
            True
            sage: (B2 * B3)^5 == (B3 * B2)^5
            True
            sage: B1 * B3 * B1 * B3 * B1 * B3 * B1 == B3 * B1 * B3 * B1 * B3 * B1 * B3
            True
        """
        data = self.coxeter_type()
        coxeter_matrix = self.coxeter_matrix()
        n = coxeter_matrix.rank()

        # Determine the base field
        if data.is_simply_laced():
            from sage.rings.integer_ring import ZZ
            base_ring = ZZ
        elif data.is_finite():
            from sage.rings.number_field.number_field import QuadraticField
            letter = data.cartan_type().type()
            if letter in ['B', 'C', 'F']:
                base_ring = QuadraticField(2)
            elif letter == 'G':
                base_ring = QuadraticField(3)
            elif letter == 'H':
                base_ring = QuadraticField(5)
            else:
                from sage.rings.universal_cyclotomic_field import UniversalCyclotomicField
                base_ring = UniversalCyclotomicField()
        else:
            from sage.rings.universal_cyclotomic_field import UniversalCyclotomicField
            base_ring = UniversalCyclotomicField()

        # Construct the matrices
        from sage.matrix.args import SparseEntry
        from sage.matrix.matrix_space import MatrixSpace
        from sage.rings.polynomial.laurent_polynomial_ring import LaurentPolynomialRing
        import sage.rings.abc
        poly_ring = LaurentPolynomialRing(base_ring, 't')
        q = -poly_ring.gen()
        MS = MatrixSpace(poly_ring, n, sparse=True)
        one = MS.one()
        # FIXME: Hack because there is no ZZ \cup \{ \infty \}: -1 represents \infty
        if isinstance(base_ring, sage.rings.abc.UniversalCyclotomicField):
            E = base_ring.gen

            def val(x):
                if x == -1:
                    return 2 * q
                elif x == 1:
                    return 1 + q**2
                else:
                    E2x = E(2 * x)
                    return q * (E2x + ~E2x)
        elif isinstance(base_ring, sage.rings.abc.NumberField_quadratic):
            from sage.rings.universal_cyclotomic_field import UniversalCyclotomicField
            E = UniversalCyclotomicField().gen

            def val(x):
                if x == -1:
                    return 2 * q
                elif x == 1:
                    return 1 + q**2
                else:
                    return q * base_ring((E(2 * x) + ~E(2 * x)).to_cyclotomic_field())
        else:
            def val(x):
                if x == -1:
                    return 2 * q
                elif x == 1:
                    return 1 + q**2
                elif x == 2:
                    return 0
                elif x == 3:
                    return q
                else:
                    from sage.functions.trig import cos
                    from sage.symbolic.constants import pi
                    return q * base_ring(2 * cos(pi / x))
        index_set = data.index_set()
        gens = [one - MS([SparseEntry(i, j, val(coxeter_matrix[index_set[i], index_set[j]]))
                          for j in range(n)])
                for i in range(n)]
        invs = [one - q**-2 * MS([SparseEntry(i, j, val(coxeter_matrix[index_set[i], index_set[j]]))
                                  for j in range(n)])
                for i in range(n)]
        return [gens, invs]

    Element = ArtinGroupElement


class FiniteTypeArtinGroup(ArtinGroup):
    r"""
    A finite-type Artin group.

    An Artin group is *finite-type* or *spherical* if the corresponding
    Coxeter group is finite. Finite type Artin groups are known to be
    torsion free, have a Garside structure given by `\Delta` (see
    :meth:`delta`) and have a center generated by `\Delta`.

    .. SEEALSO::

        :class:`ArtinGroup`

    EXAMPLES::

        sage: ArtinGroup(['E',7])                                                       # needs sage.rings.number_field
        Artin group of type ['E', 7]

    Since the word problem for finite-type Artin groups is solvable, their
    Cayley graph can be locally obtained as follows (see :issue:`16059`)::

        sage: def ball(group, radius):
        ....:     ret = set()
        ....:     ret.add(group.one())
        ....:     for length in range(1, radius):
        ....:         for w in Words(alphabet=group.gens(), length=length):
        ....:              ret.add(prod(w))
        ....:     return ret
        sage: A = ArtinGroup(['B',3])                                                   # needs sage.rings.number_field
        sage: GA = A.cayley_graph(elements=ball(A, 4), generators=A.gens()); GA         # needs sage.combinat sage.graphs sage.rings.number_field
        Digraph on 32 vertices

    Since the Artin group has nontrivial relations, this graph contains less
    vertices than the one associated to the free group (which is a tree)::

        sage: F = FreeGroup(3)
        sage: GF = F.cayley_graph(elements=ball(F, 4), generators=F.gens()); GF         # needs sage.combinat sage.graphs sage.combinat
        Digraph on 40 vertices
    """
    def delta(self):
        r"""
        Return the `\Delta` element of ``self``.

        EXAMPLES::

            sage: A = ArtinGroup(['B',3])                                               # needs sage.rings.number_field
            sage: A.delta()                                                             # needs sage.rings.number_field
            s3*(s2*s3*s1)^2*s2*s1

            sage: A = ArtinGroup(['G',2])                                               # needs sage.rings.number_field
            sage: A.delta()                                                             # needs sage.rings.number_field
            (s2*s1)^3

            sage: B = BraidGroup(5)                                                     # needs sage.combinat
            sage: B.delta()                                                             # needs sage.combinat
            s0*s1*s2*s3*s0*s1*s2*s0*s1*s0
        """
        return self._standard_lift(self._coxeter_group.long_element())

    Element = FiniteTypeArtinGroupElement



# sage_setup: distribution = sagemath-groups
"""
Type ``groups.presentation.<tab>`` to access examples
of groups implemented as finite presentations (quotients of
free groups).
"""

# groups imported here will be available
# via  groups.presentation.<tab>
#
# Do not use this file for code
#
# If you import a new group, then add an
# entry to the list in the module-level
# docstring of groups/groups_catalog.py

from .finitely_presented_named import DihedralPresentation as Dihedral
from .finitely_presented_named import CyclicPresentation as Cyclic
from .finitely_presented_named import DiCyclicPresentation as DiCyclic
from .finitely_presented_named import FinitelyGeneratedAbelianPresentation as FGAbelian
from .finitely_presented_named import FinitelyGeneratedHeisenbergPresentation as Heisenberg
from .finitely_presented_named import KleinFourPresentation as KleinFour
from .finitely_presented_named import SymmetricPresentation as Symmetric
from .finitely_presented_named import QuaternionPresentation as Quaternion
from .finitely_presented_named import AlternatingPresentation as Alternating
from .finitely_presented_named import BinaryDihedralPresentation as BinaryDihedral
from .finitely_presented_named import CactusPresentation as Cactus



# sage_setup: distribution = sagemath-groups
from sage.structure.element cimport Element, MonoidElement, MultiplicativeGroupElement

cdef class SemimonomialTransformation(MultiplicativeGroupElement):
    cdef tuple v
    cdef object perm, alpha

    cdef _new_c(self)
    cpdef _mul_(self, other)



# sage_setup: distribution = sagemath-groups
# sage.doctest: needs sage.rings.finite_rings
r"""
Semimonomial transformation group

The semimonomial transformation group of degree `n` over a ring `R` is
the semidirect product of the monomial transformation group of degree `n`
(also known as the complete monomial group over the group of units
`R^{\times}` of `R`) and the group of ring automorphisms.

The multiplication of two elements `(\phi, \pi, \alpha)(\psi, \sigma, \beta)`
with

    - `\phi, \psi \in  {R^{\times}}^n`

    - `\pi, \sigma \in S_n` (with the multiplication `\pi\sigma`
      done from left to right (like in GAP) --
      that is, `(\pi\sigma)(i) = \sigma(\pi(i))` for all `i`.)

    - `\alpha, \beta \in Aut(R)`

is defined by

.. MATH::

    (\phi, \pi, \alpha)(\psi, \sigma, \beta) =
    (\phi \cdot \psi^{\pi, \alpha}, \pi\sigma, \alpha \circ \beta)

where
`\psi^{\pi, \alpha} = (\alpha(\psi_{\pi(1)-1}), \ldots, \alpha(\psi_{\pi(n)-1}))`
and the multiplication of vectors is defined elementwisely. (The indexing
of vectors is `0`-based here, so `\psi = (\psi_0, \psi_1, \ldots, \psi_{n-1})`.)

.. TODO::

    Up to now, this group is only implemented for finite fields because of
    the limited support of automorphisms for arbitrary rings.

AUTHORS:

- Thomas Feulner (2012-11-15): initial version

EXAMPLES::

    sage: S = SemimonomialTransformationGroup(GF(4, 'a'), 4)
    sage: G = S.gens()
    sage: G[0]*G[1]
    ((a, 1, 1, 1); (1,2,3,4), Ring endomorphism of Finite Field in a of size 2^2
      Defn: a |--> a)

TESTS::

    sage: TestSuite(S).run()
    sage: TestSuite(S.an_element()).run()
"""
from __future__ import annotations

from sage.groups.group import FiniteGroup
from sage.structure.unique_representation import UniqueRepresentation
from sage.categories.action import Action
from sage.combinat.permutation import Permutation
from sage.groups.semimonomial_transformations.semimonomial_transformation import SemimonomialTransformation
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from sage.rings.integer import Integer


class SemimonomialTransformationGroup(FiniteGroup, UniqueRepresentation):
    r"""
    A semimonomial transformation group over a ring.

    The semimonomial transformation group of degree `n` over a ring `R` is
    the semidirect product of the monomial transformation group of degree `n`
    (also known as the complete monomial group over the group of units
    `R^{\times}` of `R`) and the group of ring automorphisms.

    The multiplication of two elements `(\phi, \pi, \alpha)(\psi, \sigma, \beta)`
    with

        - `\phi, \psi \in  {R^{\times}}^n`

        - `\pi, \sigma \in S_n` (with the multiplication `\pi\sigma`
          done from left to right (like in GAP) --
          that is, `(\pi\sigma)(i) = \sigma(\pi(i))` for all `i`.)

        - `\alpha, \beta \in Aut(R)`

    is defined by

    .. MATH::

        (\phi, \pi, \alpha)(\psi, \sigma, \beta) =
        (\phi \cdot \psi^{\pi, \alpha}, \pi\sigma, \alpha \circ \beta)

    where
    `\psi^{\pi, \alpha} = (\alpha(\psi_{\pi(1)-1}), \ldots, \alpha(\psi_{\pi(n)-1}))`
    and the multiplication of vectors is defined elementwisely. (The indexing
    of vectors is `0`-based here, so `\psi = (\psi_0, \psi_1, \ldots, \psi_{n-1})`.)

    .. TODO::

        Up to now, this group is only implemented for finite fields because of
        the limited support of automorphisms for arbitrary rings.

    EXAMPLES::

        sage: F.<a> = GF(9)
        sage: S = SemimonomialTransformationGroup(F, 4)
        sage: g = S(v = [2, a, 1, 2])
        sage: h = S(perm = Permutation('(1,2,3,4)'), autom=F.hom([a**3]))
        sage: g*h
        ((2, a, 1, 2); (1,2,3,4),
         Ring endomorphism of Finite Field in a of size 3^2 Defn: a |--> 2*a + 1)
        sage: h*g
        ((2*a + 1, 1, 2, 2); (1,2,3,4),
         Ring endomorphism of Finite Field in a of size 3^2 Defn: a |--> 2*a + 1)
        sage: S(g)
        ((2, a, 1, 2); (),
         Ring endomorphism of Finite Field in a of size 3^2 Defn: a |--> a)
        sage: S(1)
        ((1, 1, 1, 1); (),
         Ring endomorphism of Finite Field in a of size 3^2 Defn: a |--> a)
    """
    Element = SemimonomialTransformation

    def __init__(self, R, len):
        r"""
        Initialization.

        INPUT:

        - ``R`` -- a ring

        - ``len`` -- the degree of the monomial group

        OUTPUT: the complete semimonomial group

        EXAMPLES::

            sage: F.<a> = GF(9)
            sage: S = SemimonomialTransformationGroup(F, 4)
        """
        if not R.is_field():
            raise NotImplementedError('the ring must be a field')
        self._R = R
        self._len = len

        from sage.categories.finite_groups import FiniteGroups
        super().__init__(category=FiniteGroups())

    def _element_constructor_(self, arg1, v=None, perm=None, autom=None, check=True):
        r"""
        Coerce ``arg1`` into this permutation group, if ``arg1`` is 0,
        then we will try to coerce ``(v, perm, autom)``.

        INPUT:

        - ``arg1`` -- (optional) either the integers 0, 1 or an element of ``self``

        - ``v`` -- (optional) a vector of length ``self.degree()``

        - ``perm`` -- (optional) a permutation of degree ``self.degree()``

        - ``autom`` -- (optional) an automorphism of the ring

        EXAMPLES::

            sage: F.<a> = GF(9)
            sage: S = SemimonomialTransformationGroup(F, 4)
            sage: S(1)
            ((1, 1, 1, 1); (), Ring endomorphism of Finite Field in a of size 3^2 Defn: a |--> a)
            sage: g = S(v=[1,1,1,a])
            sage: S(g)
            ((1, 1, 1, a); (), Ring endomorphism of Finite Field in a of size 3^2 Defn: a |--> a)
            sage: S(perm=Permutation('(1,2)(3,4)'))
            ((1, 1, 1, 1); (1,2)(3,4), Ring endomorphism of Finite Field in a of size 3^2 Defn: a |--> a)
            sage: S(autom=F.hom([a**3]))
            ((1, 1, 1, 1); (), Ring endomorphism of Finite Field in a of size 3^2 Defn: a |--> 2*a + 1)
        """
        from sage.categories.homset import End
        R = self.base_ring()
        if arg1 == 0:
            if v is None:
                v = [R.one()] * self.degree()
            if perm is None:
                perm = Permutation(range(1, self.degree() + 1))
            if autom is None:
                autom = R.hom(R.gens())

            if check:
                try:
                    v = [R(x) for x in v]
                except TypeError:
                    raise TypeError('the vector attribute %s ' % v +
                                    'should be iterable')
                if len(v) != self.degree():
                    raise ValueError('the length of the vector is %s,' % len(v) +
                                     ' should be %s' % self.degree())
                if not all(x.parent() is R and x.is_unit() for x in v):
                    raise ValueError('there is at least one element in the ' +
                                     'list %s not lying in %s ' % (v, R) +
                                     'or which is not invertible')
                try:
                    perm = Permutation(perm)
                except TypeError:
                    raise TypeError('the permutation attribute %s ' % perm +
                                    'could not be converted to a permutation')
                if len(perm) != self.degree():
                    txt = 'the permutation length is {}, should be {}'
                    raise ValueError(txt.format(len(perm), self.degree()))

                try:
                    if autom.parent() != End(R):
                        autom = End(R)(autom)
                except TypeError:
                    raise TypeError('%s of type %s' % (autom, type(autom)) +
                                    ' is not coerceable to an automorphism')
            return self.Element(self, v, perm, autom)
        else:
            try:
                if arg1.parent() is self:
                    return arg1
            except AttributeError:
                pass
            try:
                from sage.rings.integer import Integer
                if Integer(arg1) == 1:
                    return self()
            except TypeError:
                pass
            raise TypeError('the first argument must be an integer' +
                            ' or an element of this group')

    def base_ring(self):
        r"""
        Return the underlying ring of ``self``.

        EXAMPLES::

            sage: F.<a> = GF(4)
            sage: SemimonomialTransformationGroup(F, 3).base_ring() is F
            True
        """
        return self._R

    def degree(self) -> Integer:
        r"""
        Return the degree of ``self``.

        EXAMPLES::

            sage: F.<a> = GF(4)
            sage: SemimonomialTransformationGroup(F, 3).degree()
            3
        """
        return self._len

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        EXAMPLES::

            sage: F.<a> = GF(4)
            sage: SemimonomialTransformationGroup(F, 3).an_element()  # indirect doctest
            ((a, 1, 1); (1,3,2), Ring endomorphism of Finite Field in a of size 2^2 Defn: a |--> a + 1)
        """
        R = self.base_ring()
        v = [R.primitive_element()] + [R.one()] * (self.degree() - 1)
        p = Permutation([self.degree()] + list(range(1, self.degree())))

        if not R.is_prime_field():
            f = R.hom([R.gen()**R.characteristic()])
        else:
            f = R.Hom(R).identity()
        return self(0, v, p, f)

    def __contains__(self, item) -> bool:
        r"""
        EXAMPLES::

            sage: F.<a> = GF(4)
            sage: S = SemimonomialTransformationGroup(F, 3)
            sage: 1 in S  # indirect doctest
            True
            sage: a in S  # indirect doctest
            False
        """
        try:
            self(item, check=True)
        except TypeError:
            return False
        return True

    def gens(self) -> tuple:
        r"""
        Return a tuple of generators of ``self``.

        EXAMPLES::

            sage: F.<a> = GF(4)
            sage: SemimonomialTransformationGroup(F, 3).gens()
            (((a, 1, 1); (),
              Ring endomorphism of Finite Field in a of size 2^2 Defn: a |--> a),
             ((1, 1, 1); (1,2,3),
              Ring endomorphism of Finite Field in a of size 2^2 Defn: a |--> a),
             ((1, 1, 1); (1,2),
              Ring endomorphism of Finite Field in a of size 2^2 Defn: a |--> a),
             ((1, 1, 1); (),
              Ring endomorphism of Finite Field in a of size 2^2 Defn: a |--> a + 1))
        """
        from sage.groups.perm_gps.permgroup_named import SymmetricGroup
        R = self.base_ring()
        l = [self(v=([R.primitive_element()] + [R.one()] * (self.degree() - 1)))]
        l.extend(self(perm=Permutation(g))
                 for g in SymmetricGroup(self.degree()).gens())
        if R.is_field() and not R.is_prime_field():
            l.append(self(autom=R.hom([R.primitive_element()**R.characteristic()])))
        return tuple(l)

    def order(self) -> Integer:
        r"""
        Return the number of elements of ``self``.

        EXAMPLES::

            sage: F.<a> = GF(4)
            sage: SemimonomialTransformationGroup(F, 5).order() == (4-1)**5 * factorial(5) * 2
            True
        """
        from sage.arith.misc import factorial
        from sage.categories.homset import End
        n = self.degree()
        R = self.base_ring()
        if R.is_field():
            multgroup_size = len(R) - 1
            autgroup_size = R.degree()
        else:
            multgroup_size = R.unit_group_order()
            autgroup_size = len([x for x in End(R) if x.is_injective()])
        return multgroup_size**n * factorial(n) * autgroup_size

    def _get_action_(self, X, op, self_on_left):
        r"""
        If ``self`` is the semimonomial group of degree `n` over `R`, then
        there is the natural action on `R^n` and on matrices `R^{m \times n}`
        for arbitrary integers `m` from the left.

        See also:
        :class:`~sage.groups.semimonomial_transformations.semimonomial_transformation_group.SemimonomialActionVec` and
        :class:`~sage.groups.semimonomial_transformations.semimonomial_transformation_group.SemimonomialActionMat`

        EXAMPLES::

            sage: F.<a> = GF(4)
            sage: s = SemimonomialTransformationGroup(F, 3).an_element()
            sage: v = (F**3).0
            sage: s*v   # indirect doctest
            (0, 1, 0)
            sage: M = MatrixSpace(F, 3).one()
            sage: s*M   # indirect doctest
            [    0     1     0]
            [    0     0     1]
            [a + 1     0     0]
        """
        if self_on_left:
            try:
                A = SemimonomialActionVec(self, X)
                return A
            except ValueError:
                pass

            try:
                A = SemimonomialActionMat(self, X)
                return A
            except ValueError:
                pass

        return None

    def _repr_(self) -> str:
        r"""
        Return a string describing ``self``.

        EXAMPLES::

            sage: F.<a> = GF(4)
            sage: SemimonomialTransformationGroup(F, 3)  # indirect doctest
            Semimonomial transformation group over Finite Field in a of size 2^2 of degree 3
        """
        return ('Semimonomial transformation group over %s' % self.base_ring() +
                ' of degree %s' % self.degree())

    def _latex_(self) -> str:
        r"""
        Method for describing ``self`` in LaTeX.

        EXAMPLES::

            sage: F.<a> = GF(4)
            sage: latex(SemimonomialTransformationGroup(F, 3))  # indirect doctest
            \left(\Bold{F}_{2^{2}}^3\wr\langle (1,2,3), (1,2) \rangle \right) \rtimes \operatorname{Aut}(\Bold{F}_{2^{2}})
        """
        from sage.groups.perm_gps.permgroup_named import SymmetricGroup
        ring_latex = self.base_ring()._latex_()
        return ('\\left(' + ring_latex + '^' + str(self.degree()) + '\\wr' +
                SymmetricGroup(self.degree())._latex_() +
                ' \\right) \\rtimes \\operatorname{Aut}(' + ring_latex + ')')


class SemimonomialActionVec(Action):
    r"""
    The natural left action of the semimonomial group on vectors.

    The action is defined by:
    `(\phi, \pi, \alpha)*(v_0, \ldots, v_{n-1}) :=
    (\alpha(v_{\pi(1)-1}) \cdot \phi_0^{-1}, \ldots, \alpha(v_{\pi(n)-1}) \cdot \phi_{n-1}^{-1})`.
    (The indexing of vectors is `0`-based here, so
    `\psi = (\psi_0, \psi_1, \ldots, \psi_{n-1})`.)
    """
    def __init__(self, G, V, check=True):
        r"""
        Initialization.

        EXAMPLES::

            sage: F.<a> = GF(4)
            sage: s = SemimonomialTransformationGroup(F, 3).an_element()
            sage: v = (F**3).1
            sage: s*v   # indirect doctest
            (0, 0, 1)
        """
        if check:
            from sage.modules.free_module import FreeModule_generic
            if not isinstance(G, SemimonomialTransformationGroup):
                raise ValueError('%s is not a semimonomial group' % G)
            if not isinstance(V, FreeModule_generic):
                raise ValueError('%s is not a free module' % V)
            if V.ambient_module() != V:
                raise ValueError('%s is not equal to its ambient module' % V)
            if V.dimension() != G.degree():
                raise ValueError('%s has a dimension different to the degree of %s' % (V, G))
            if V.base_ring() != G.base_ring():
                raise ValueError('%s and %s have different base rings' % (V, G))

        Action.__init__(self, G, V.dense_module())

    def _act_(self, a, b):
        r"""
        Apply the semimonomial group element `a` to the vector `b`.

        EXAMPLES::

            sage: F.<a> = GF(4)
            sage: s = SemimonomialTransformationGroup(F, 3).an_element()
            sage: v = (F**3).1
            sage: s*v   # indirect doctest
            (0, 0, 1)
        """
        b = b.apply_map(a.get_autom())
        b = self.codomain()(a.get_perm().action(b))
        return b.pairwise_product(self.codomain()(a.get_v_inverse()))


class SemimonomialActionMat(Action):
    r"""
    The left action of
    :class:`~sage.groups.semimonomial_transformations.semimonomial_transformation_group.SemimonomialTransformationGroup`
    on matrices over the same ring whose number of columns is equal to the degree.
    See :class:`~sage.groups.semimonomial_transformations.semimonomial_transformation_group.SemimonomialActionVec`
    for the definition of the action on the row vectors of such a matrix.
    """
    def __init__(self, G, M, check=True):
        r"""
        Initialization.

        EXAMPLES::

            sage: F.<a> = GF(4)
            sage: s = SemimonomialTransformationGroup(F, 3).an_element()
            sage: M = MatrixSpace(F, 3).one()
            sage: s*M  # indirect doctest
            [    0     1     0]
            [    0     0     1]
            [a + 1     0     0]
        """
        if check:
            from sage.matrix.matrix_space import MatrixSpace
            if not isinstance(G, SemimonomialTransformationGroup):
                raise ValueError('%s is not a semimonomial group' % G)
            if not isinstance(M, MatrixSpace):
                raise ValueError('%s is not a matrix space' % M)
            if M.ncols() != G.degree():
                raise ValueError('the number of columns of %s' % M +
                                 ' and the degree of %s are different' % G)
            if M.base_ring() != G.base_ring():
                raise ValueError('%s and %s have different base rings' % (M, G))
        Action.__init__(self, G, M)

    def _act_(self, a, b):
        r"""
        Apply the semimonomial group element `a` to the matrix `b`.

        EXAMPLES::

            sage: F.<a> = GF(4)
            sage: s = SemimonomialTransformationGroup(F, 3).an_element()
            sage: M = MatrixSpace(F, 3).one()
            sage: s*M  # indirect doctest
            [    0     1     0]
            [    0     0     1]
            [a + 1     0     0]
        """
        return self.codomain()([a * x for x in b.rows()])



# sage_setup: distribution = sagemath-groups
r"""
Kernel Subgroups

The kernel of a homomorphism implemented as a subgroup.

AUTHORS:

- Travis Scrimshaw (1-2023): Initial version
"""

# ****************************************************************************
#       Copyright (C) 2023 Travis Scrimshaw <tcscrims at gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.categories.groups import Groups
from sage.structure.element_wrapper import ElementWrapper
from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation


class KernelSubgroup(UniqueRepresentation, Parent):
    r"""
    The kernel (normal) subgroup.

    Let `\phi : G \to H` be a group homomorphism. The kernel
    `K = \{\phi(g) = 1 | g \in G\}` is a normal subgroup of `G`.
    """
    def __init__(self, morphism):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: S2 = SymmetricGroup(2)
            sage: S3 = SymmetricGroup(3)
            sage: H = Hom(S3, S2)
            sage: phi = H(S2.__call__)
            sage: from sage.groups.kernel_subgroup import KernelSubgroup
            sage: K = KernelSubgroup(phi)
            sage: TestSuite(K).run()
        """
        self._morphism = morphism
        cat = Groups().Subobjects()
        base_cat = morphism.domain().category()
        if base_cat in Groups().Finite():
            cat = cat.Finite()
        elif base_cat in Groups().Enumerated():
            cat = cat.Enumerated()
        Parent.__init__(self, category=cat)

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: S2 = SymmetricGroup(2)
            sage: S3 = SymmetricGroup(3)
            sage: H = Hom(S3, S2)
            sage: phi = H(S2.__call__)
            sage: from sage.groups.kernel_subgroup import KernelSubgroup
            sage: KernelSubgroup(phi)
            Kernel subgroup defined by Generic morphism:
              From: Symmetric group of order 3! as a permutation group
              To:   Symmetric group of order 2! as a permutation group
        """
        return "Kernel subgroup defined by {}".format(self._morphism)

    def gens(self) -> tuple:
        r"""
        Return the generators of ``self``.

        EXAMPLES::

            sage: S2 = SymmetricGroup(2)
            sage: S3 = SymmetricGroup(3)
            sage: H = Hom(S3, S2)
            sage: phi = H(S2.__call__)
            sage: from sage.groups.kernel_subgroup import KernelSubgroup
            sage: K = KernelSubgroup(phi)
            sage: K.gens()
            ((),)
        """
        if self.ambient() in Groups().Finite():
            return tuple(self)
        raise NotImplementedError("only implemented for finite groups")

    def defining_morphism(self):
        r"""
        Return the defining morphism of ``self``.

        EXAMPLES::

            sage: PJ3 = groups.misc.PureCactus(3)                                       # needs sage.rings.number_field
            sage: PJ3.defining_morphism()                                               # needs sage.rings.number_field
            Conversion via _from_cactus_group_element map:
              From: Cactus Group with 3 fruit
              To:   Symmetric group of order 3! as a permutation group
        """
        return self._morphism

    @cached_method
    def ambient(self):
        r"""
        Return the ambient group of ``self``.

        EXAMPLES::

            sage: PJ3 = groups.misc.PureCactus(3)                                       # needs sage.rings.number_field
            sage: PJ3.ambient()                                                         # needs sage.rings.number_field
            Cactus Group with 3 fruit
        """
        return self._morphism.domain()

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        EXAMPLES::

            sage: PJ3 = groups.misc.PureCactus(3)                                       # needs sage.rings.number_field
            sage: PJ3.an_element()                                                      # needs sage.rings.number_field
            1
        """
        return self.element_class(self, self.ambient().one())

    def lift(self, x):
        r"""
        Lift ``x`` to the ambient group of ``self``.

        EXAMPLES::

            sage: PJ3 = groups.misc.PureCactus(3)                                       # needs sage.rings.number_field
            sage: PJ3.lift(PJ3.an_element()).parent()                                   # needs sage.rings.number_field
            Cactus Group with 3 fruit
        """
        return x.value

    def retract(self, x):
        r"""
        Convert ``x`` to an element of ``self``.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: J3 = groups.misc.Cactus(3)
            sage: s12,s13,s23 = J3.group_generators()
            sage: PJ3 = groups.misc.PureCactus(3)
            sage: elt = PJ3.retract(s23*s12*s23*s13); elt
            s[2,3]*s[1,2]*s[2,3]*s[1,3]
            sage: elt.parent() is PJ3
            True
        """
        return self._element_constructor_(x)

    def _element_constructor_(self, x):
        r"""
        Construct an element of ``self`` from ``x``.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: J3 = groups.misc.Cactus(3)
            sage: s12,s13,s23 = J3.group_generators()
            sage: PJ3 = groups.misc.PureCactus(3)
            sage: elt = PJ3(s23*s12*s23*s13)
            sage: elt.parent() is PJ3
            True
        """
        if self._morphism(x) != self._morphism.codomain().one():
            raise ValueError("{} is not in the kernel of {}".format(x, self._morphism))
        return self.element_class(self, x)

    def __iter__(self):
        r"""
        Iterate through ``self``.

        EXAMPLES::

            sage: S2 = SymmetricGroup(2)
            sage: S3 = SymmetricGroup(3)
            sage: H = Hom(S3, S2)
            sage: phi = H(S2.__call__)
            sage: from sage.groups.kernel_subgroup import KernelSubgroup
            sage: K = KernelSubgroup(phi)
            sage: list(K)
            [()]
        """
        for g in self.ambient():
            try:
                yield self(g)
            except ValueError:
                pass

    class Element(ElementWrapper):
        def _mul_(self, other):
            r"""
            Multiply ``self`` and ``other``.

            EXAMPLES::

                sage: # needs sage.rings.number_field
                sage: J3 = groups.misc.Cactus(3)
                sage: s12,s13,s23 = J3.group_generators()
                sage: PJ3 = groups.misc.PureCactus(3)
                sage: elt = PJ3(s23*s12*s23*s13)
                sage: elt * elt
                s[2,3]*s[1,2]*s[2,3]*s[1,2]*s[2,3]*s[1,2]
            """
            return type(self)(self.parent(), self.value * other.value)

        def __invert__(self):
            r"""
            Return the inverse of ``self``.

            EXAMPLES::

                sage: # needs sage.rings.number_field
                sage: J3 = groups.misc.Cactus(3)
                sage: s12,s13,s23 = J3.group_generators()
                sage: PJ3 = groups.misc.PureCactus(3)
                sage: elt = PJ3(s23*s12*s23*s13)
                sage: ~elt
                s[1,2]*s[2,3]*s[1,2]*s[1,3]
            """
            return type(self)(self.parent(), ~self.value)
